from .statistical_methods import (find_statistical_saboteurs,
                                  statistics_report)
from .logical_methods import (design_test_batch, find_logical_saboteurs,
                              plot_batch,
                              generate_batch_report,
                              generate_combinatorial_groups)
from .tools import csv_to_groups_data


from saboteurs import find_logical_saboteurs

# NOTE: the teo definitions below could also be read from a CSV:
# groups, failed_groups = csv_to_groups_data('find_logical_saboteurs.csv')

groups = {
    1: ['A', 'C', 'D'],
    2: ['B', 'C', 'E'],
    3: ['A', 'B', 'D'],
    4: ['D', 'F', 'G']
}
failed_groups = [2, 4]

result = find_logical_saboteurs(groups, failed_groups=failed_groups)
print (result)


from collections import OrderedDict
import itertools
import numpy as np
from .minimal_cover import minimal_cover


def generate_combinatorial_groups(elements_per_position, prefix='group_'):
    """Generate all possibilities from a combinatorial design.
    
    Parameters
    ----------
    elements_per_position
      A dict {position_name: [elements at this position]}
    prefix
      Prefix to generate names for the generated groups, which will be of the
      form prefix_001, prefix_002, etc.

    Returns
    -------
      groups
        An ordered dict {group_001: [e1, e2, e3, e4], group_002: [...] etc.}
        where each group contains exactly one element from each position, and
        all possible combinations of elements are there.
    """
    if hasattr(elements_per_position, 'values'):
        elements_per_position = list(elements_per_position.values())
    groups = list(itertools.product(*elements_per_position))
    n_zeros = int(np.log10(len(groups))) + 1
    return OrderedDict([
        (prefix + str(i + 1).zfill(n_zeros), group)
        for i, group in enumerate(groups)
    ])

def _minimal_elements_group_coverage(groups):
    elements_group_coverage = {}
    for group_name, elements in groups.items():
        for element in elements:
            if element not in elements_group_coverage:
                elements_group_coverage[element] = set()
            elements_group_coverage[element].add(group_name)
    return minimal_cover(set(groups.keys()), elements_group_coverage.items())

def design_test_batch(possible_groups, max_saboteurs=1):
    """Select a subset of the groups that enables to identify bad elements.

    Parameters
    ----------
    possible_groups
      A dict of the form {group_name: [elements in group]}.
    
    max_saboteurs
      The maximum number of potential bad elements among all elements in all
      the groups. A bad element is an element which will make every group
      that contains it "fail"

    Returns
    -------
    selected_groups, error
      A subset of carefully selected elements of ``possible_groups`` as a dict
      {group: [elements in group]}. The selected groups are such that
      knowing which of them failed or succeeded will be enough information
      to identify all bad elements in the original ``possible_groups`` set.
      If there is an error, then selected groups is [] and the error is a
      string explaining what went wrong.

    """
    covering_elements = _minimal_elements_group_coverage(possible_groups)
    lcov = len(covering_elements)
    if lcov <= max_saboteurs:
        return None, (
            "Not possible to detect up to %d saboteurs, as the following %d "
            "elements collectively cover all constructs, and therefore all "
            "constructs would fail at once if these %d elements alone were"
            "saboteurs:\n\n%s\n\n. Remove these elements from the problem or "
            " decrease the max number of saboteurs if possible.") % (
                max_saboteurs, lcov, lcov, ", ".join(covering_elements))
    all_elements = set(
        element
        for elements in possible_groups.values()
        for element in elements
    )
    
    product = itertools.product(*((1 + max_saboteurs) * (all_elements,)))
    all_tuples = set(
        tuple_ for tuple_ in product 
        if len(set(tuple_)) == len(tuple_)
    )
    
    def all_x_without_ys(group):
        return set(
            tuple_ for tuple_ in all_tuples
            if (tuple_[0] in group)
            and not any((e in group) for e in tuple_[1:])
        )
    
    x_without_ys_sets = [
        (name, all_x_without_ys(group))
        for name, group in possible_groups.items()
    ]
    selected = minimal_cover(all_tuples, x_without_ys_sets)
    if selected is None:
        return [], 'No solution found.'
    keys = list(possible_groups.keys())
    selected = sorted(selected, key=lambda group: keys.index(group))
    return OrderedDict((g, possible_groups[g]) for g in selected), None

def _groups_fail_table(groups):
    """Returns a dict associating each element to the groups it can fail.
    This function is used both in ``plot_elements_in_group`` and
    ``find_logical_saboteurs``.
    
    Parameters
    ----------
    groups
      Ordered dict of the form {group_name: [elements in groups]}

    Returns
    -------
    fail_table
      A dict {element_name: list[True False False ...]} where list[i] is True
      if the element is part of the i-th group in the provided ordered
      dictionnary ``groups``.
    """
    groups = list(groups.values())
    all_elements = []
    for group in groups:
        for element in group:
            if element not in all_elements:
                all_elements.append(element)
    return OrderedDict([
        (element, [(element in group) for group in groups])
        for element in all_elements
    ])

def find_logical_saboteurs(groups, failed_groups):
    """Identify bad and suspicious elements from groups failure data
    
    Parameters
    ----------
    groups
      A dict {group_name: [elements in that group]}
    failed_groups
      A list [group_name_1, group_name_2, ...] of the names of all groups that
      experimentally failed.


    Returns
    -------
    {'saboteurs': [...], 'suspicious': []}
      Where ``suspicious`` is the list of all elements which do not appear in
      successful group, and ``saboteurs`` is the list of suspicious elements
      which are also the only suspicious element in at least one group.
    """
    groups = OrderedDict(groups.items())
    failed_groups = set(failed_groups)
    groups_list = np.array(list(groups.keys()))
    fail_table = {
        element: set(groups_list[groups_])
        for element, groups_ in _groups_fail_table(groups).items()
    }
    suspicious = set(
        element for element, element_groups in fail_table.items()
        if element_groups < failed_groups
    )
    confirmed = set(
        suspect_element
        for suspect_element in suspicious
        if len(fail_table[suspect_element].difference(set().union(*(
            fail_table[other_suspect]
            for other_suspect in suspicious.difference({suspect_element})
        ))))
    )
    return dict(saboteurs=list(confirmed),
                suspicious=list(suspicious.difference(confirmed)))


from .logical_methods import (design_test_batch,
                              find_logical_saboteurs,
                              generate_combinatorial_groups)
from .reports import  plot_batch, generate_batch_report


from collections import OrderedDict
import numpy as np
import matplotlib.pyplot as plt
import flametree
from .logical_methods import _groups_fail_table

def plot_batch(groups, ax=None):
    """Plot a diagram of all groups and the elements they contain.

    The ``groups`` parameter is a dict {group_name: [elements in the group]}.
    The ax is a Matplotlib Ax object on which to plot. If none is provided a
    new ax will be created and returned at the end.
    """
    groups = OrderedDict(groups.items())
    
    fail_table = _groups_fail_table(groups)
    array = np.array(list(fail_table.values())).T[::-1]
    lines, cols = len(array), len(array[0])

    if ax is None:
        _, ax = plt.subplots(1)
    ax.imshow(array, cmap='Purples')
    ax.set_aspect('equal')
    for y in range(lines):
        ax.axhline(y + .5, c='white')
    ax.set_yticks(range(lines))
    for x in range(cols):
        ax.axvline(x + .5, c='white')
    ax.set_xticks(range(cols))
    ax.set_xticklabels(list(fail_table.keys()), rotation=90)
    ax.set_yticklabels(list(groups.keys())[::-1])
    ax.set_xlim(-.5, cols - .5)
    ax.set_ylim(-.5, lines - .5)
    ax.set_aspect('equal')
    return ax

def generate_batch_report(groups, target='@memory', group_naming='group',
                          plot_format='pdf'):
    """Generate a report with CSV and plot describing a groups batch.
    
    Parameters
    ----------
    groups
      A (ordered) dict {group_name: [elements in the group]}.
    
    target
      Either path to a folder, or a zip file, or "@memory" to return raw
      data of a zip file containing the report.
    
    group_naming
      Word that will replace "group" in the report, e.g. "assembly", "team",
      etc.
    
    plot_format
      Formal of the plot (pdf, png, jpeg, etc.)

    """
    root = flametree.file_tree(target)
    csv = ("%s,elements\n" % group_naming) + "\n".join([
        ",".join([group] + list(elements))
        for group, elements in groups.items()
    ])
    root._file('%ss.csv' % group_naming).write(csv)
    ax = plot_batch(groups)
    ax.set_title('Elements per %s' % group_naming)
    ax.figure.savefig(
        root._file('%ss.%s' % (group_naming, plot_format)).open('wb'),
        bbox_inches='tight', format=plot_format)
    return root._close()


from saboteurs import (generate_combinatorial_groups,
                       design_test_batch,
                       generate_batch_report)

elements_per_position = {
    "Position_1": ['A', 'B', 'C'],
    "Position_2": ['D', 'E', 'F', 'G'],
    "Position_3": ['H', 'I', 'J', 'K'],
    "Position_4": ['L', 'M', 'N'],
}
possible_groups = generate_combinatorial_groups(elements_per_position,
                                                prefix='G')
selected_groups, error = design_test_batch(possible_groups, max_saboteurs=2)
generate_batch_report(selected_groups, plot_format='png',
                      target='design_test_batch_report')


from scipy.stats import binom
from collections import OrderedDict
from sklearn import linear_model, metrics
from sklearn.feature_selection import SelectFpr, f_classif
import numpy as np
from copy import deepcopy

def _find_twins(groups_data, almost_twins_threshold=0.8):
    groups_members_list = [data["members"] for data in groups_data.values()]
    all_members = set([
        member for l in groups_members_list
        for member in l
    ])
    profiles = {
        member: [member in parts_list for parts_list in groups_members_list]
        for member in all_members
    }
    profiles = {
        member: profile
        for member, profile in profiles.items()
        if min(profile) != max(profile)
    }
    all_members = sorted(set(profiles.keys()))
    twins = {}
    almost_tweens = {m: set() for m in all_members}
    has_tweens = {m: False for m in all_members}
    for i, m1 in enumerate(all_members):
        if has_tweens[m1]:
            continue
        for m2 in all_members[i + 1:]:
            if has_tweens[m2]:
                continue
            corr = np.corrcoef(profiles[m1], profiles[m2])[1, 0]
            if corr > 0.999:
                if m1 not in twins:
                    twins[m1] = set()
                twins[m1].add(m2)
                has_tweens[m1] = has_tweens[m2] = True
            elif corr > almost_twins_threshold:
                almost_tweens[m1].add((m2, corr))
                almost_tweens[m2].add((m1, corr))
    return twins, almost_tweens, has_tweens

def find_statistical_saboteurs(groups_data, pvalue_threshold=0.1, effect_threshold=0,
                   max_significant_members=10):
    """Return statistics on possible bad elements in the data.

    Parameters
    ----------
    groups_data
      Result of ``csv_to_groups_data()``

    pvalue_threshold
      Only failure-associated elements with a p-value below this threshold
      will be included in the final statistics

    """
    groups_data = deepcopy(groups_data)
    twins, almost_tweens, has_twins = _find_twins(groups_data)
    members_sets = [set(group['members']) for group in groups_data.values()]
    all_members = set().union(*members_sets)
    conserved_members = members_sets[0].intersection(*members_sets)
    members_with_twins = set().union(*twins.values())
    varying_members = sorted(all_members.difference(conserved_members)
                                        .difference(members_with_twins))

    # Build the data

    def build_data_and_observed(selected_members, by_group=False):
        data = []
        observed = []
        for group_name, group_data in groups_data.items():
            attempts = int(group_data['attempts'])
            failures = int(group_data['failures'])
            vector = [[
                (mb in group_data['members'])
                for mb in selected_members
            ]]
            if by_group:
                data += vector
                observed.append(1.0 * failures / attempts)
            else:
                data += attempts * vector
                observed += (attempts - failures) * [0] + failures * [1]
        return np.array(data), np.array(observed)

    # LASSO model (gives positive / negative impact)
    data, observed = build_data_and_observed(varying_members)
    regression = linear_model.RidgeCV()
    regression.fit(data, observed)

    # ANOVA analysis (for p-values)
    selector = SelectFpr(f_classif, alpha=pvalue_threshold)
    selector.fit(data, observed)

    # select the most interesting parts
    data_ = zip(selector.pvalues_, regression.coef_, varying_members)
    significant_members = OrderedDict([
        (name, {'pvalue': pvalue, 'twins': twins.get(name, [])})
        for pvalue, coef, name in sorted(data_)
        if (pvalue < pvalue_threshold) and (coef > 0)
    ])

    if len(significant_members) == 0:
        return {
            'groups_data': groups_data,
            'conserved_members': conserved_members,
            'varying_members': varying_members,
            'significant_members': significant_members,
        }
    # LASSO model (significant parts only)
    data, observed = build_data_and_observed(significant_members)
    regression.fit(data, observed)
    zipped = zip(regression.coef_, significant_members.items())
    for coef, (name, data_) in zipped:
        data_['effect'] = coef
    for member in list(significant_members.keys()):
        if significant_members[member]["effect"] < effect_threshold:
            significant_members.pop(member)

    # print (significant_members)
    # significant_members = significant_members[:max_significant_members]

    # Build a classifier to compute a L1 score
    classifier = linear_model.LogisticRegressionCV(penalty='l2')
    classifier.fit(data, observed)
    f1_score = metrics.f1_score(observed, classifier.predict(data))

    # Find constructs which are less explained by the parts:
    data, observed = build_data_and_observed(significant_members,
                                             by_group=True)
    regression.fit(data, observed)
    predictions = regression.predict(data)
    zipped = zip(groups_data.values(), observed, predictions)
    intercept = min(0.9, max(0.1, regression.intercept_))
    for group_data, obs, pred in zipped:
        std = binom.std(group_data['attempts'], intercept) / group_data['attempts']
        group_data['failure_rate'] = obs
        group_data['deviation'] = np.round((obs - pred) / std, decimals=1)

    return {
        'groups_data': groups_data,
        'conserved_members': conserved_members,
        'varying_members': varying_members,
        'significant_members': significant_members,
        'f1_score': f1_score
    }



from saboteurs import csv_to_groups_data, find_statistical_saboteurs, statistics_report
groups_data = csv_to_groups_data("assemblies_data.csv")
analysis_results = find_statistical_saboteurs(groups_data)
statistics_report(analysis_results, "genetic_parts_example.pdf", replacements=[
    ('groups', 'assemblies'),
    ('group', 'assembly'),
    ('member', 'part')
])



import os
import string
from collections import OrderedDict

import pandas
import numpy as np
from pdf_reports import pug_to_html, write_report

from ..version import __version__

THIS_PATH = os.path.dirname(os.path.realpath(__file__))
STYLESHEET = os.path.join(THIS_PATH, 'assets', 'report_style.css')
TEMPLATE = os.path.join(THIS_PATH, 'assets', 'report_template.pug')

def saboteurs_pug_to_html(**context):
    defaults = {
        'logo_url': os.path.join(THIS_PATH, 'assets', 'logo.png'),
        'sidebar_text': "Generated by Saboteurs version " + __version__
    }
    for k in defaults:
        if k not in context:
            context[k] = defaults[k]
    return pug_to_html(TEMPLATE, **context)

def make_groups_table(analysis_results):
    """Return a Pandas dataframe indicating with elements belong to each group.
    """
    return pandas.DataFrame.from_records([
        OrderedDict([
            ('Group', group),
            ("Failure Rate (%)", int(100 * group_data['failure_rate']))] + [
                (member, "✔" * (member in group_data['members']))
                for member in analysis_results['significant_members']
            ] + [("Mystery", "nan" if (str(group_data['deviation']) == "nan")
                 else "*" * abs(max(0, int(group_data['deviation']))))]
        )
        for (group, group_data) in analysis_results['groups_data'].items()
    ]).sort_values('Failure Rate (%)', ascending=False)

def make_members_table(analysis_results):
    """Return a Pandas dataframe with significance/impact of the main elements.
    """
    return pandas.DataFrame.from_records([
        OrderedDict([
            ('Member', member),
            ("p-value", "%.03f" % np.round(data['pvalue'], 3)),
            ("Effect", "+%d%%" % (100 * data['effect'])),
            ("Twins", "<br/>".join(data["twins"]))
        ])
        for (member, data) in analysis_results['significant_members'].items()
    ]).sort_values('p-value')

def replace_in_text(text, replacements, capitalize=True):
    """Return the text with all the replacements done.

    Parameters
    ----------

    text
      Some string.

    replacements
      A list of the form ``[("text_to_replace", "text_replacing"), ...]``

    capitalize
      If true, then for each pair ``"text to replace"=>"text replacing"`` in
      ``replacements``, the couple ``"Text to replace"=>"Text replacing"``
      is also added (note the capitalization of the first letter)
    """
    if capitalize:
        replacements = list(replacements) + [
            (string.capwords(target), string.capwords(replacement))
            for target, replacement in replacements
        ]
    for target, replacement in replacements:
        text = text.replace(target, replacement)
    return text

def statistics_report(analysis_results, outfile, replacements=()):
    """Procude a PDF reports from the results of ``find_statistical_saboteurs()``.

    Parameters
    ----------

    analysis_results
      The result of ``saboteurs.find_statistical_saboteurs``.

    outfile
      Path to the final PDF file, or file-like object, or '@memory' to
      return binary data of the PDF report.

    replacements
      A list of the form ``[("text_to_replace", "text_replacing"), ...]``
    """
    if len(analysis_results['significant_members']) == 0:
        html = saboteurs_pug_to_html(
            members_table=None,
            groups_table=None,
            f1_score=None,
            saboteurs_found=False
        )
    else:
        html = saboteurs_pug_to_html(
            members_table=make_members_table(analysis_results),
            groups_table=make_groups_table(analysis_results),
            f1_score=analysis_results['f1_score'],
            saboteurs_found=True
        )
    html = replace_in_text(html, capitalize=True, replacements=replacements)
    return write_report(html, outfile, extra_stylesheets=(STYLESHEET,))



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open('saboteurs/version.py').read()) # loads __version__

setup(
    name='saboteurs',
    version=__version__,
    author='Zulko',

    description='Identify agents impairing success accross experiments.',
    long_description=open('pypi-readme.rst').read(),
    license='see LICENSE.txt',
    url='https://github.com/Edinburgh-Genome-Foundry/saboteurs',
    keywords="statistics weakest link DNA part validation",
    packages=find_packages(exclude='docs'),
    include_package_data=True,
    install_requires=["numpy", "pandas", "scipy", "sklearn", "pdf_reports",
                      "matplotlib", "flametree"])



from saboteurs import csv_to_groups_data, find_statistical_saboteurs, statistics_report
groups_data = csv_to_groups_data("basic_example.csv")
analysis_results = find_statistical_saboteurs(groups_data)
statistics_report(analysis_results, "basic_example.pdf")
print ("Done. See basic_example.pdf for a report.")


from collections import OrderedDict

def csv_to_groups_data(csv_path=None, csv_string=None):
    """Read a CSV to get the data to feed to ``find_statistical_saboteurs()``
    or ``find_logical_saboteurs()``.

    See examples of such a file in the code repository:

    https://github.com/Edinburgh-Genome-Foundry/saboteurs/

    Returns
    -------
    groups, failed_groups
      For datasheets for logical saboteur finding.


    group_data
      For datasheets for statistical saboteur finding. The data is of the form
      
      >>> {"Exp. 1": {
      >>>      exp_id: "Exp. 1",
      >>>     attempts: 7,
      >>>     failures: 10,
      >>>     members: ["Alice", "Bob"]}
      >>>  }
      >>>  "Exp. 2": { etc...
    """
    if csv_string is None:
         with open(csv_path, 'r') as f:
             csv_string = f.read()
    lines = [[e.strip() for e in l.split(',') if len(e.strip())]
                for l in csv_string.split('\n') if len(l)]
    groups = OrderedDict([])
    if 'result' in lines[0]:
        failed_groups = []
        for line in lines[1:]:
            name, result, members = line[0], line[1], line[2:]
            groups[name] = members
            if result != 'success':
                failed_groups.append(name)
        return groups, failed_groups
    else:
        for line in lines[1:]:
            (name, attempts, failures), members = line[:3], line[3:]
            groups[name] = dict(id=name, attempts=int(attempts),
                                failures=int(failures), members=members)
        return groups